function [p,dXdx,pnums] = MeshSensGen3d(fFunc,ppre,tet,pnums,fH,endtol,dpParams,varargin)
%MeshSensGen 3-D Mesh Sensitivities Generator using Distance Functions.
%   [p,t,dXdx] = MeshSensGen2d(fFunc,ppre,t,pnums,fH,endtol,dpParams,varargin))
%
%      Outputs:
%      p:         Node coordinates (Nx2)
%      t:         Triangle indices (NTx3)
%      dXdx:      Sensitivity matrix (2NxParams)
%      
%      Inputs:
%      ppre:      Pre-mesh node coordinates (Nx2) (x,y)
%      t:         Triangle indices (NTx3)
%      fFunc:     Boundary functionset
%          fFunc.Params: Control parameters (Nx1)
%          fFunc.F(p):   Boundary function (value)
%          fFunc.dF(p):  Boundary function first derivitive (2x1)
%          fFunc.ddF(p): Boundary function second derivitive (2,2)
%          fFunc.pfix:   Matrix of fixed nodes(Nx2)
%          fFunc.PLOT:   Matrix of nodes used to plot the boundary(x,y)
%
%      pnums:     Node numbers (of p) corresponding to row fixed nodes in fFunc.pfix  (Nx1)
%      fH:        Edge length function h(p)
%      endtol:    Termination tolerence for Newton step (default 1e-8)
%      dpParam:   Perturbation values corresponding to control variables fFunc.Params
%   Copyright (C) 2020 Campbell Bam. See COPYRIGHT.TXT for details.

%% 1 Intialisation and parsing of inputs
%% 1.1 Describe each bar by a unique pair of nodes
p = ppre; % set working nodes p equal to input nodes pin
bars = [tet(:,[1,2]);tet(:,[1,3]);tet(:,[2,3]);tet(:,[1,4]);tet(:,[2,4]);tet(:,[3,4])];   % Extract truss node pairs sortrows([reshape(x(:,e(:,1)),[],1) reshape(x(:,e(:,2)),[],1)])
bars = unique(sort(bars,2),'rows'); % Remove duplicate bars

%% 1.2 Obtain node numbers corresponding to the fixed nodes
% iuf pnums is not given, then the nodes are determined to be the closest
% nodes in the domain 
if nargin<4 || isempty(pnums)
    % Determine fixed node indexes
    pfix = fFunc.pfix(p); pnums = zeros(size(pfix,1),1);
    for i = 1:size(pfix,1)
        pdists = sum([pfix(i,1)-p(:,1), pfix(i,2)-p(:,2), pfix(i,3)-p(:,3) ].^2,2).^0.5; %calculate distances 
        pnums(i,1) = find(pdists == min(pdists));
    end
end
p(pnums,:) = fFunc.pfix(p); %Ensure the fixed nodes match those in fFunc.pfix(p)

%% 1.3 Determine MPC nodes
AllFaces = [tet(:,[1 2 3]);tet(:,[1 2 4]);tet(:,[1 3 4]);tet(:,[2 3 4])];
[UniqueFaces,~,ic] = unique(sort(AllFaces,2),'rows');
SurfaceFaceNos = find(histc(ic,1:max(ic)) == 1); % Find unique Faces (mesh surface)
SurfacePs = unique(reshape(UniqueFaces(SurfaceFaceNos,:),{},1)); % Extract mesh edge nodes
MPCnums = setdiff(SurfacePs,pnums); % MPC node numbers

% pnums = [pnums; MPCnums(5:end)]; MPCnums = setdiff(MPCnums,pnums);

%% 1.4 Establish free nodes (not MPC or fixed)
fnums = setdiff((1:size(p,1))',[MPCnums; pnums]); % establish free nodes

%% 1.5 Determine DOFs
nmpc = size(MPCnums,1); % Number of MPC nodes therefore equations
ndof = numel(p); % no. of nodal DOFs
pdof = sort([pnums*3-2;pnums*3-1;pnums*3]); % Prescribed DOFs
MPCdof = sort([MPCnums*3-2;MPCnums*3-1;MPCnums*3]); % MPC DOFs
fdof = sort([fnums*3-2;fnums*3-1;fnums*3]); % Free DOFs
% ldof = ((ndof+1):(ndof+nmpc))'; % lambda DOFs
lambda = zeros(0,1); % Initialise zero lambda vector

%% 1.6 Handle potentially missing input variables
if nargin <5 | isempty(fH); fhFlag = 1; % Treats p,t final desired sizes
else, fhFlag = 0; end % This will use the desired length function fH

% Sets the Newton method convergence tolerence if none was given
if nargin <6 | isempty(endtol); endtol = 1e-8; end

% Sets the parameter purturbation vector to 1e-8 if none was given
if nargin <7 | isempty(dpParams)
    dpParams = 1e-8*ones(size(fFunc.Params,1),1);
elseif all(size(dpParams)==1)
    dpParams = dpParams*ones(size(fFunc.Params,1),1);
end

%% 1.7 Initialisation values
PlotOn = true; % Ploton = true mesh will be plotted during updates
allowNewtSteps = 3; % Allow certain number of newton steps
newtonsteps = inf; % Newton Step Counter,

%% 2. Obtain mesh equilibrium using Newton's method
while 1
    % 3.1 Calculate the currect bar lengths
    barvec=p(bars(:,1),:) - p(bars(:,2),:);  % List of bar vectors
    L=sqrt(sum(barvec.^2,2));                % L = Current bar bar lengths
    
    % 3.2 Calculate desired truss lengths L0
    if newtonsteps >= allowNewtSteps && fhFlag == 0
        % fhFlag = 0 implies that a distance function is expected to be
        % used.
        % Recalculate bar lengths every "allowed newton step", this allows
        % the quadratic convergence behaviour of the newtons method to take
        % effect  
        hbars=feval(fH,(p(bars(:,1),:)+p(bars(:,2),:))/2,varargin{:}); % Actual desired lengths according to fH
        L0 = hbars;
        newtonsteps = 0; % Reset 
        allowNewtSteps = allowNewtSteps + 1; % Increase allowed Newton steps to avoid convergence failure associated with changing lengths L0
    elseif fhFlag == 1 % Set desired length L0 = current Length L
        L0 = L; % Set desired lengths L0
        lambda = zeros(0,1);; % Reset lambda, since objective lengths L0 reset
        fhFlag = 2; % Flag set to skip over using setting edge lengths
    end
    
    %% 3.3. Truss Consistant tangent contributions
    F = zeros(ndof,1); dFdX = sparse(ndof,ndof); % initialise arrays
    ivec = zeros(size(bars,1)*16,1);jvec = ivec; vvec = ivec;vecloc = 0;%%% Initialise vectors for speed (used to speed up dFdX(dof,dof)=dFdX(dof,dof) + dFdX)
    
    for i = 1:size(bars,1)
        dof = [bars(i,1)*3-2, bars(i,1)*3-1,bars(i,1)*3, bars(i,2)*3-2, bars(i,2)*3-1,bars(i,2)*3]'; % Obtain the DOFs associated with the bar
        x = [p(bars(i,1),:),p(bars(i,2),:)]'; % Obtain the set of nodal coordinates asscoiated with p1 and p2 of bar
        [Ft,dFdX_lcl,broken] = Truss3D(x,L0(i,1)); % Calculate force and consistent tangent contributions
        F(dof,1) = F(dof) + Ft; % Assemble force vector
        % dFdX(dof,dof) = dFdX(dof,dof) + dFdX_lcl; % Assemble consistent tangent
        % end
        
        %%% % Vectorised assembly implimentation signifcantly faster for
        %%% % sparse matrices assembly
        %%% % replaces dFdX(dof,dof) = dFdX(dof,dof) + dFdX;
        for ibas = 1:size(dof,1) %%%
            for jbas = 1:size(dof,1) %%%
                vecloc = vecloc + 1; %%%
                ivec(vecloc,1) = dof(ibas); %%%
                jvec(vecloc,1) = dof(jbas); %%%
                vvec(vecloc,1) = dFdX_lcl(ibas,jbas); %%%
            end %%%
        end %%%
        %%%
    end%%%
    dFdX = sparse(ivec,jvec,vvec); %%% replaces sparse indexing of dFdX(dof,dof)
    %%% end vectorised implimentation
    %% 4. Langrangian MPC contributions
    % Intialise vectors
    dFmpcdXc = sparse(nmpc,ndof); 
    ddFmpcddXcLam = sparse(ndof,ndof); % note size matches dFdX
    Fmpc = zeros(0,1);
    % 4.1
    % Calculate and assemble MPC contributions
    nLambda =0;
    Doublejoint = zeros(0,1);
    lambdaowner = zeros(0,1);
    for i = 1:size(MPCnums,1)
        dof = [MPCnums(i,1)*3-2:MPCnums(i,1)*3]; % Get global DOFs
        
        lclF = fFunc.F(p(MPCnums(i,1),:));
        lcldF  = fFunc.dF(p(MPCnums(i,1),:));
        lclddF = fFunc.ddF(p(MPCnums(i,1),:));
        if length(lclF)>1
           Doublejoint = [Doublejoint; i];
        end
        
        for j = 1: length(lclF)
            nLambda = nLambda + 1;
            lambdaowner = [lambdaowner;i];
            if length(lambda) < nLambda
               lambda(nLambda,1) = 0;
            end
            Fmpc(nLambda,1) = lclF{j};  % Calculate residual of node
            dFmpcdXc(nLambda,dof) = lcldF{j}; % Calculate and assemble residual of node
             % Calculate consistent tangent contribution
            ddFmpcddXcLam(dof,dof) = ddFmpcddXcLam(dof,dof)+lclddF{j}*lambda(nLambda,1);% Assemble
        end

    end

    nmpc = length(lambda); ldof = ((ndof+1):(ndof+nmpc))';
    % 4.2. Assemble Linear System
    % Combine truss and MPC components of consistent tangent
    ConsTan = [(dFdX+ddFmpcddXcLam) dFmpcdXc'
                dFmpcdXc            sparse(nmpc,nmpc)];

    % 4.3 Combine truss and MPC components of RHS
    RHS = zeros(ndof+nmpc,1);
    RHS([pdof;fdof],1) = F([pdof;fdof],1);
    RHS(MPCdof,1) = F(MPCdof,1)+(dFmpcdXc(:,MPCdof))'*lambda;
    RHS(ldof,1) = Fmpc;
    
    %% 5 Solve Newton step
    solvedof = [fdof;MPCdof;ldof]; % DOFs for use in Newton update solution (excludes pdofs)
    

    % Newton update
    Supdate = zeros(ndof+nmpc,1);
    Supdate(solvedof) = ConsTan(solvedof,solvedof)\(-RHS(solvedof)); % Obtainupdate
    
    % Update limiter (prefereable for if fH exists, reduces risk of
    % nodes snapping through). It however requires  tweaking to obtain
    % adequate performance from the code. The smaller the
    % UpdateLimiter, the less likely a node will snap through.
%             dp = [Supdate(1:3:ndof,1), Supdate(2:3:ndof,1),Supdate(3:3:ndof,1)]; % delta p vec
%             barchange = sum((dp(bars(:,1),:)-dp(bars(:,2),:)).^2,2).^0.5;
%             maxchange = max(abs(barchange./L));
%             UpdateLimiter = 0.01;
%             if maxchange > UpdateLimiter
%                 delta = UpdateLimiter/maxchange;
%             else
%                 delta = 1;
%             end
%             Supdate = delta.*Supdate;
        
    p = p+real([Supdate(1:3:ndof),Supdate(2:3:ndof),Supdate(3:3:ndof)]); % Update nodes p (extract the nodal coordinate X)
    lambda = lambda + Supdate(ldof); % Update the lambda
    
    if PlotOn == true
        %PlotMPCs points around edge
        MESHplotter(p,tet,1);hold on;
        xlabel('x')
        ylabel('y')
        zlabel('z')
        plot3(p(MPCnums,1),p(MPCnums,2),p(MPCnums,3),'r.');% Plot MPC nodes
        plot3(p(pnums,1),p(pnums,2),p(pnums,3),'k*'); % Plot fixed nodes
        axis equal;
        drawnow,hold off;
    end
    
    %% 6. Termination criterion:
    Normupdt = norm(Supdate); % Obtain norm of the update vector
    disp(['Update norm: ' num2str(Normupdt)]);
    newtonsteps = newtonsteps +1; % Increment Newton counter (since last L0 calc)
    if (Normupdt < endtol)
        break
    end
end

%% 7. Obtain sensitivities
% Iterate through the parameters
dRHS = zeros(ndof+nmpc,size(fFunc.Params,1));
for ParNum = 1:size(fFunc.Params,1)
    deps = dpParams(ParNum); % Extract perturbation value associated with control parameter
    dFunc = fFunc; % Create temporary boundary
    dFunc.Params(ParNum,1) = fFunc.Params(ParNum,1) + deps; % Perturb control parameter to have sensitivities calculated
    
    dXs = zeros(ndof,1); % Initialise a peturbed nodal coordinate vector
    
    %% 7.1. Add the Fixed point updates w.r.t. param updates
    StablePDOFs = reshape([pnums*3-2 pnums*3-1 pnums*2],[],1); % Keep a stable version of the pdofs that correlates to fFunc.pfix(p)
    dXs(StablePDOFs,1) = reshape(dFunc.pfix(p) - fFunc.pfix(p),[],1); % Calculate and extract the update components of pfix to an update vector
    dXpdxi(:,ParNum) = dXs(pdof,1)./deps; % Complete the sensitivity calculation for the pdofs
    
    %% 7.2. MPC contributions for perturbed paramters
    PtdFmpcdXc = sparse(nmpc,ndof); % intialise 
    PtFmpc = zeros(nmpc,1); % initialise
    nLambda = 0;
    for i = 1:size(MPCnums,1)
        dof = [MPCnums(i,1)*3-2:MPCnums(i,1)*3]; % Get global DOFs
        lclF = dFunc.F(p(MPCnums(i,1),:));
        lcldF  = dFunc.dF(p(MPCnums(i,1),:));
        for j = 1: length(lclF)
            nLambda = nLambda + 1;
            PtFmpc(nLambda,1) = lclF{j};  % Calculate residual of node
            PtdFmpcdXc(nLambda,dof) = lcldF{j}; % Calculate and assemble residual of node
        end
    end
    
    % 7.3. Assemble RHS components for sensitivity calculation
    dRHS(ldof,ParNum) = (PtFmpc-Fmpc)./deps; % Numerical sensitivity of dFmpcdxi (Fmpc actually equal to zero and equation reduces to PtFmpc/deps Fmpc shown for numerical sensitivity
    dRHS(MPCdof,ParNum) = (PtdFmpcdXc(:,MPCdof)-dFmpcdXc(:,MPCdof))'*lambda./deps; % Numerical sensitivity of ddFmpcdXcdxi
end

% 8. Solve sensitivities
dSupdate = zeros(ndof+nmpc,size(fFunc.Params,1));
dSupdate(solvedof,:) = ConsTan(solvedof,solvedof)\(-dRHS(solvedof,:) - ConsTan(solvedof,pdof)*dXpdxi); % note change due to sensitivity cal supdate no longer nega tive

% 9.. Assemble into sensitivity matrix dXdx
dXdx([fdof;MPCdof;pdof],:) = [dSupdate([fdof;MPCdof],:); dXpdxi];

MESHplotter(p,tet,1);

end
